-- Boilerplate:
import algebra.algebra.basic
import algebra.big_operators.basic
import data.equiv.ring
namespace examples
open_locale big_operators

/-
Morphisms and subobjects in mathlib
===================================
Or: program to the class, not the type

Mathlib has all kinds of morphisms:
 * (functions)
 * additive, multiplicative, monoid, ring, algebra homomorphisms
 * linear maps
 * multiplicative, ring isomorphisms
 * linear equivalences
 * continuous maps
 * monotone maps / order homomorphisms
 * order isomorphisms
 * homeomorphisms
 * ...


Similarly, mathlib has all kinds of subobjects:
 * (subsets)
 * (additive) submonoids, (additive) subgroups, subsemirings, subrings, subfields
 * submodules
 * subalgebras, intermediate fields
 * closed subspaces
 * ...

There are actually a few ways you can define morphisms and subobjects,
it's not obvious which one is the best,
and it's important to get the design right because they find use throughout mathlib.
So I'll go through the three main designs we use, pointing out their advantages and disadvantages.

Most examples will be about morphisms:
the design for subobjects can be exactly the same if you replace the function
`f : α → β` with a subset `s : set α` (or a function `f : α → Prop`).
For example, `map_add f x y : f (x + y) = f x + f y` corresponds to
`add_mem s x y : x ∈ s → y ∈ s → (x + y) ∈ s`

Some design criteria:
 * natural notation: if `f` is a ring hom, can we write `map_add f x y : f (x + y) = f x + f y`?

                     can we write `map_add (2 *) x y : 2 * (x + y) = 2 * x + 2 * y`?

                     if we declare `f` as a monoid hom, can we easily upgrade it to a ring hom?

 * less bookkeeping: can we write `map_add (f ∘ 2 • g)` instead of
    `map_add (is_ring_hom.comp f_is_ring_hom (smul_is_ring_hom 2 g_is_ring_hom)`?

 * works with automation: can tactics like `simp` automatically apply `map_add`?

                          can we avoid complicated higher-order matching?

 * supports rewriting: if we have `∀ x, f x = g x`, and `f` is a ring hom,
                       how much work is it to say `g` is a ring hom?

 * no boilerplate: if `f` is a monotone ring hom, can we write `map_add f`
                   instead of `map_add f.to_ring_hom.to_add_monoid_hom.to_add_hom`?

                   when defining monotone ring homs, can we avoid copying over lemmas
                   `map_add`, `map_zero`, ..., `map_finset_sum`, `map_determinant`?

 * structure on morphisms: can we define a module structure on linear endomorphisms?

I'll discuss three designs for morphisms that exist(ed) in mathlib,
focussing on the algebraic parts of the library (which I'm most familiar with).

-/

/-
Design 1: unbundled morphisms

This is the original design from the beginning of mathlib.
To say `f` is a ring homomorphism, we have a function `f : R → S` and a
proof `hf : is_ring_hom f`.

The `is_ring_hom` predicate states `f` preserves the ring operations `*`, `+`, `1` and `0`.
We write it as a class to provide a bit of automation.
-/
-- Adapted from `deprecated/group.lean:71`
class is_add_monoid_hom {M N : Type*} [add_monoid M] [add_monoid N] (f : M → N) : Prop :=
(map_add [] : ∀ x y : M, f (x + y) = f x + f y)
(map_zero [] : f 0 = 0)

-- Adapted from `deprecated/group.lean:77`
class is_monoid_hom {M N : Type*} [monoid M] [monoid N] (f : M → N) : Prop :=
(map_mul [] : ∀ x y : M, f (x * y) = f x * f y)
(map_one [] : f 1 = 1)

-- Adapted from `deprecated/ring.lean:64`
class is_ring_hom {R S : Type*} [semiring R] [semiring S] (f : R → S)
  extends is_monoid_hom f, is_add_monoid_hom f

#check is_add_monoid_hom.map_add

/- Unbundled predicates mean there is no easy way to treat the collection of all
ring homs `R → R` as a monoid in its own right. -/

/- Notation is very natural, since everything the user sees is just a function. -/

-- Adapted from `deprecated/group.lean:166`
instance mul.is_add_monoid_hom {R : Type*} [ring R] (c : R) :
  is_add_monoid_hom ((*) c) :=
{ map_zero := mul_zero c,
  map_add := mul_add c }

example {R : Type*} [ring R] (c a b : R) : c * (a + b) = c * a + c * b :=
is_add_monoid_hom.map_add ((*) c) a b

/-
Lean's typeclass system provides a database of known ring homomorphisms.
This means you can write `map_add (id R)`, and Lean can supply the implicit
`id.is_ring_hom R` argument to `map_add`.
-/
-- Adapted from `deprecated/ring.lean:93`
instance id.is_ring_hom (R : Type*) [semiring R] :
  is_ring_hom (id : R → R) :=
{ map_mul := λ _ _, rfl,
  map_one := rfl,
  map_add := λ _ _, rfl,
  map_zero := rfl }

/-
However, finding these instances is difficult beyond the most trivial cases:
we cannot automate the fact that composing two ring homs is a ring hom.
-/
-- Adapted from `deprecated/ring.lean:97`
lemma comp.is_ring_hom {R S T : Type*} (f : R → S) (g : S → T)
  [semiring R] [semiring S] [semiring T] [hf : is_ring_hom f] [hg : is_ring_hom g] :
  is_ring_hom (g ∘ f) :=
{ map_add := λ x y, by simp [hf.map_add]; rw hg.map_add; refl,
  map_mul := λ x y, by simp [hf.map_mul]; rw hg.map_mul; refl,
  map_one := by simp [hf.map_one]; exact hg.map_one,
  map_zero := by simp [hf.map_zero]; exact hg.map_zero }

section

/-
Trying to add this fact to the database of class instances causes timeouts:
-/
attribute [instance] comp.is_ring_hom

set_option trace.class_instances true
example {R S : Type*} [semiring R] :
  is_ring_hom (id : R → R) := by apply_instance

/-
What's going on is Lean tries to infer `is_ring_hom id` by applying `comp.is_ring_hom`,
getting a new goal `is_ring_hom (id ∘ id)`, applying `comp.is_ring_hom`,
resulting in a new goal `is_ring_hom (id ∘ id ∘ id)`, etc.
-/

end

/-
Automation doesn't work very well either because `simp` doesn't do (much) higher-order matching:
-/
example {R S : Type*} (f : R → S) [semiring R] [semiring S] [hf : is_ring_hom f] (x y : R) :
  f (x + y) = f x + f y := by simp only [is_add_monoid_hom.map_add]

/-
In unbundled morphisms' favour, typeclass inheritance means there is very little boilerplate.
Lemmas on parent classes are immediately available:
-/
example {R S : Type*} (f : R → S) [semiring R] [semiring S] [hf : is_ring_hom f] (x y : R) :
  f (x * y) = f x * f y := is_monoid_hom.map_mul f x y

namespace bundled
/-
The second design used bundled morphisms, where we package the function
and proofs into a single tuple.
Each algebraic structure gets its own type of morphisms:
-/
-- Adapted from `algebra/group/hom.lean:248`
structure monoid_hom (M N : Type*) [monoid M] [monoid N] :=
(to_fun : M → N)
(map_mul : ∀ x y, to_fun (x * y) = to_fun x * to_fun y)
(map_one : to_fun 1 = 1)

-- Adapted from `algebra/ring/basic.lean:329`
structure ring_hom (R S : Type*) [semiring R] [semiring S]
  extends monoid_hom R S :=
(map_add : ∀ x y, to_fun (x + y) = to_fun x + to_fun y)
(map_zero : to_fun 0 = 0)

/-
The `has_coe_to_fun` typeclass allows us to keep using `f (x + y)` notation
even if `f` does not have a function type:
-/
-- Adapted from `algebra/group/hom.lean:406`
instance monoid_hom.has_coe_to_fun (M N : Type*) [monoid M] [monoid N] :
  has_coe_to_fun (monoid_hom M N) (λ _, M → N) :=
{ coe := monoid_hom.to_fun }

example {M N : Type*} [monoid M] [monoid N] (f : monoid_hom M N) (x y : M) :
  f (x * y) = f x * f y := f.map_mul x y

-- is syntactic sugar for
example {M N : Type*} [monoid M] [monoid N] (f : monoid_hom M N) (x y : M) :
  (coe_fn f : M → N) (x * y) = (coe_fn f : M → N) x * (coe_fn f : M → N) y := f.map_mul x y

-- which expands to
example {M N : Type*} [monoid M] [monoid N] (f : monoid_hom M N) (x y : M) :
  f.to_fun (x * y) = f.to_fun x * f.to_fun y := f.map_mul x y

/-
Notation becomes slightly more cumbersome since we need to specify
whether `id` refers to the function, or monoid hom, or ring hom, ...
-/

-- Adapted from `algebra/group/hom.lean:661`
def monoid_hom.id (M : Type*) [monoid M] : monoid_hom M M :=
{ to_fun := id,
  map_one := rfl,
  map_mul := λ _ _, rfl }

def ring_hom.id (R : Type*) [semiring R] : ring_hom R R :=
{ to_fun := id,
  map_one := rfl,
  map_mul := λ _ _, rfl,
  map_zero := rfl,
  map_add := λ _ _, rfl }

-- Adapted from `algebra/group/hom.lean:688`
def monoid_hom.comp {M N O : Type*} [monoid M] [monoid N] [monoid O]
  (f : monoid_hom M N) (g : monoid_hom N O) : monoid_hom M O :=
{ to_fun := g ∘ f,
  map_one := by simp [coe_fn, has_coe_to_fun.coe, monoid_hom.has_coe_to_fun, monoid_hom.map_one],
  map_mul := by simp [coe_fn, has_coe_to_fun.coe, monoid_hom.has_coe_to_fun, monoid_hom.map_mul] }
/-
We also have to write `add_monoid_hom.mul_left c` instead of `((*) c)`.
-/
-- Adapted from `algebra/ring/basic.lean:303`
def add_monoid_hom.mul_left {R : Type*} [ring R] (c : R) :
  add_monoid_hom R R :=
{ to_fun := (*) c,
  map_zero' := mul_zero c,
  map_add' := mul_add c }

end bundled

/-
The bundled homs work a lot better with automation and rewrites:
-/
example {R S : Type*} [ring R] [ring S] (f : ring_hom R S) (x y : R) :
  f (x + y) = f x + f y := by simp only [ring_hom.map_add]

example {R S : Type*} [ring R] [ring S] (f : ring_hom R S) (g : R → S) (x y : R)
  (h : (f : R → S) = g) : f (x + y) = g x + g y :=
by rw [ring_hom.map_add, h]

/-
The bookkeeping is still done nicely for us, now through the types of the variables
rather than instances on the elements.
-/

/-
The other drawback, apart from annoying notation, is the big amount of boilerplate.
If we prove that a `monoid_hom` preserves finite products,
we'll have to copy that proof over to all other bundled morphism types extending `monoid_hom`.

Thus mathlib ended up with many copies of lemmas such as `map_prod`:
-/
-- Adapted from `algebra/big_operators/basic.lean:111`
lemma monoid_hom.map_prod {ι M N : Type*} (s : finset ι) (f : ι → M) [comm_monoid M] [comm_monoid N]
  (g : monoid_hom M N) :
  g (∏ i in s, f i) = ∏ i in s, g (f i) :=
by simp only [finset.prod_eq_multiset_prod, g.map_multiset_prod, multiset.map_map]

-- Adapted from `algebra/big_operators/basic.lean:144`
lemma ring_hom.map_prod {ι R S : Type*} (s : finset ι) (f : ι → R) [comm_semiring R] [comm_semiring S]
  (g : ring_hom R S) :
  g (∏ i in s, f i) = ∏ i in s, g (f i) :=
monoid_hom.map_prod s f g.to_monoid_hom

-- Adapted from `algebra/big_operators/basic.lean:116`
lemma mul_equiv.map_prod {ι M N : Type*} (s : finset ι) (f : ι → M) [comm_monoid M] [comm_monoid N]
  (g : mul_equiv M N) :
  g (∏ i in s, f i) = ∏ i in s, g (f i) :=
monoid_hom.map_prod s f g.to_monoid_hom

-- Adapted from `data/equiv/ring.lean:386`
lemma ring_equiv.map_prod {ι R S : Type*} (s : finset ι) (f : ι → R) [comm_semiring R] [comm_semiring S]
  (g : ring_equiv R S) :
  g (∏ i in s, f i) = ∏ i in s, g (f i) :=
monoid_hom.map_prod s f g.to_monoid_hom

-- Adapted from `algebra/algebra/basic.lean:704`
lemma alg_hom.map_prod {ι R A B : Type*} (s : finset ι) (f : ι → A) [comm_semiring R]
  [comm_semiring A] [comm_semiring B] [algebra R A] [algebra R B]
  (g : alg_hom R A B) :
  g (∏ i in s, f i) = ∏ i in s, g (f i) :=
ring_hom.map_prod s f g.to_ring_hom

-- Adapted from `algebra/algebra/basic.lean:1142`
lemma alg_equiv.map_prod {ι R A B : Type*} (s : finset ι) (f : ι → A) [comm_semiring R]
  [comm_semiring A] [comm_semiring B] [algebra R A] [algebra R B]
  (g : alg_equiv R A B) :
  g (∏ i in s, f i) = ∏ i in s, g (f i) :=
g.to_alg_hom.map_prod f s

/-
In fact, we get a multiplicative amount of boilerplate:
a duplicate lemma for each operation preserved by `monoid_hom` times each subtype of `monoid_hom`.
Even worse is that this was done ad hoc and incompletely in mathlib,
so half of the time you have to go back and copy the lemma over to your favourite morphism.
-/

/-
I decided to fix this boilerplate issue using the power of typeclasses.
The insight is that the `map_prod` lemma assumes `f` has a specific type `monoid_hom`,
where all subtypes of `monoid_hom` would work just the same.
So we need to make a generic interface for all types extending `monoid_hom`.

My first step in introducing this interface was a typeclass `fun_like`
for *types* of bundled (dependent) functions,
based on Eric Wieser's `set_like` class for types of bundled subobjects.
(https://github.com/leanprover-community/mathlib/pull/6768)
-/
-- Adapted from `library/init/coe.lean:54`
class has_coe_to_fun' (F : Type*) (α : out_param (F → Type*)) :=
(coe : Π x : F, α x)

-- Adapted from `data/fun_like/basic.lean:127`
class fun_like (F : Type*)
  (α : out_param Type*) (β : out_param (α → Type*))
  extends has_coe_to_fun F (λ _, Π a : α, β a) :=
(coe_injective' : function.injective to_has_coe_to_fun.coe)

-- A typical instance looks like:
instance monoid_hom.fun_like {M N : Type*} [monoid M] [monoid N] :
  fun_like (monoid_hom M N) M (λ _, N) :=
{ coe := monoid_hom.to_fun,
  coe_injective' := λ f g h, by { cases f, cases g, congr' } }

/-
Now we can already get rid of some boilerplate
such as extensionality and congruence lemmas.
(After all, `α → β` is the type of morphisms that preserve the operation `a = b`.)
-/
#check fun_like.congr
#check fun_like.ext_iff

/-
The second step is to make a subclass of `fun_like` for all subtypes
of the type of monoid homomorphisms.
-/
-- Adapted from `algebra/group/hom.lean:260`
class monoid_hom_class (F : Type*) (M N : out_param Type*)
  [monoid M] [monoid N] extends fun_like F M (λ _, N) :=
(map_one : ∀ (f : F), f 1 = 1)
(map_mul : ∀ (f : F) (x y : M), f (x * y) = f x * f y)

/-
We keep the bundled `monoid_hom` type, but connect it to the class:
-/
-- Adapted from `algebra/group/hom.lean:265`
instance monoid_hom.monoid_hom_class {M N : Type*} [monoid M] [monoid N] :
  monoid_hom_class (monoid_hom M N) M N :=
{ coe := monoid_hom.to_fun,
  coe_injective' := λ f g h, by { cases f, cases g, congr' },
  map_one := monoid_hom.map_one',
  map_mul := monoid_hom.map_mul' }

/-
For example, ring homs are also monoid homs,
so the corresponding class becomes a subclass of `monoid_hom_class`:
-/
-- Adapted from `algebra/ring/basic.lean:354`
class ring_hom_class (F : Type*) (R S : out_param Type*)
  [semiring R] [semiring S]
  extends monoid_hom_class F R S :=
(map_zero : ∀ (f : F), f 0 = 0)
(map_add : ∀ (f : F) (x y : R), f (x + y) = f x + f y)

-- Adapted from `algebra/ring/basic.lean:382`
instance ring_hom.ring_hom_class {R S : Type*} [semiring R] [semiring S] :
  ring_hom_class (ring_hom R S) R S :=
{ coe := ring_hom.to_fun,
  coe_injective' := λ f g h, by { cases f, cases g, congr' },
  map_mul := ring_hom.map_mul',
  map_one := ring_hom.map_one',
  map_add := ring_hom.map_add',
  map_zero := ring_hom.map_zero' }

/-
Now lemmas can be made generic by parametrising over all the types of bundled morphisms,
reducing the multiplicative amount of lemmas to an additive amount.
each extension of `monoid_hom` should get a `monoid_hom_class` instance,
and each operation preserved by `monoid_hom`s should get a lemma taking a `monoid_hom_class` parameter.
-/
lemma map_prod {ι M N : Type*} (s : finset ι) (f : ι → M) [comm_monoid M] [comm_monoid N]
  {G : Type*} [monoid_hom_class G M N] (g : G) :
  g (∏ i in s, f i) = ∏ i in s, g (f i) :=
begin
  letI := classical.dec_eq ι,
  refine finset.induction_on s _ _,
  { simp [monoid_hom_class.map_one] },
  { intros i s hi ih,
    simp [finset.prod_insert hi, monoid_hom_class.map_mul, ih] },
end

/-
So we have solved one part of the boilerplate issue:
the lemmas now generalize automatically.

Boilerplate is not totally gone though:
standard maps such as `id` still have a copy for each bundled morphism type:
-/
#check monoid_hom.id
#check ring_hom.id
#check alg_hom.id
#check linear_map.id
-- etc.

/-
All of these `id`s then also have lemmas about composition with `id`, applying `id`,
the inverse (if it's a bundled isomorphism), etc.

I'm currently trying to figure out if there is a neat way to deduplicate this.
For example, a `has_lawful_id` class, or perhaps bundling the map
along with a proof that it is the identity function on elements
into a new value `id : identity (ring_hom R R)`.
-/

end examples
